import asyncio
import logging
import threading
import time

from .pilarerror import PilarError


log = logging.getLogger(__name__)


class PilarCommand(object):
    def __init__(self, command):
        self.command = command
        self.id = None
        self.time = None
        self.sent = False
        self.acknowledged = False
        self.completed = False
        self.error = None
        self.data = None
        self.values = {}

    def __call__(self, transport):
        # send command
        cmd = str(self.id) + ' ' + self.command + '\n'
        transport.write(bytes(cmd, 'utf-8'))

        # store current time and set as sent
        self.time = time.time()
        self.sent = True

    def parse(self, line):
        # split line and check ID
        s = line.split()
        if int(s[0]) != self.id:
            return

        # acknowledge
        if 'COMMAND OK' in line or 'COMMAND ERROR' in line:
            self.acknowledged = True
            if 'COMMAND ERROR' in line:
                self.error = line

        # payload
        if 'DATA INLINE' in line:
            # get key and value
            pos = line.find('=')
            key = line[line.find('DATA INLINE') + 12:pos]
            value = line[pos+1:]
            # type?
            if value[0] == '"' and value[-1] == '"':
                value = value[1:-1]
            # store it
            self.values[key] = value
            # we always store the last result as data, makes it easier for commands requesting only a single value
            self.data = value

        # finish
        elif 'COMMAND COMPLETE' in line or 'COMMAND FAILED' in line:
            self.completed = True

    def wait(self, timeout: int = 5, abort_event: threading.Event = None):
        """Wait for the command to finish.

        Args:
            timeout: Timeout for waiting in seconds.
            abort_event: When set, wait is aborted.
        """

        # wait for data
        while not self.completed:
            # sleep a little
            if abort_event is not None:
                abort_event.wait(0.1)

                # abort event set?
                if abort_event.is_set():
                    return

            else:
                time.sleep(0.1)

            # timeout reached?
            if time.time() - self.time > timeout:
                raise TimeoutError('Command took too long to execute.')


class PilarClientProtocol(asyncio.Protocol):
    """ asyncio.Protocol implementation for the Pilar interface. """

    def __init__(self, driver, loop, username, password):
        """ Creates a SicamTcpClientProtocol.
        :param driver: SicamTcpDriver instance.
        :param loop: asyncio event loop.
        :param username: Username for login.
        :param password: Password for login
        :return:
        """

        # init some stuff
        self._driver = driver
        self._buffer = ''
        self._loop = loop
        self._transport = None
        self._username = username
        self._password = password
        self._logged_in = False
        self._id = 0
        self._commands = []

        # store self in driver
        self._driver.protocol = self

    @property
    def logged_in(self):
        return self._logged_in

    def connection_made(self, transport):
        """ Called, when the protocol is connected to the server.
        :param transport: Transport connected to server.
        :return:
        """

        # store transport
        self._transport = transport

    @asyncio.coroutine
    def stop(self):
        """ Disconnect gracefully. """

        # send disconnect
        self._transport.write(b'disconnect')

        # disconnect
        self._transport.close()
        self._loop.stop()

    def data_received(self, data):
        """ Called, when new data arrives from the server.
        :param data: New chunk of data.
        :return:
        """

        # add data to buffer
        self._buffer += data.decode('utf-8')

        # create as many packets as possible
        while self._buffer and '\n' in self._buffer:
            # extract line from buffer
            length = self._buffer.find('\n')
            line = self._buffer[:length]
            self._buffer = self._buffer[length+1:]

            # AUTH?
            if 'AUTH PLAIN' in line:
                log.info('Logging into Pilar...')
                # send AUTH line
                auth = 'AUTH PLAIN "' + self._username + '" "' \
                       + self._password + '"\n'
                self._transport.write(bytes(auth, 'utf-8'))

            elif 'AUTH OK' in line:
                log.info('Authentication for Pilar successful.')
                self._logged_in = True

            elif 'AUTH FAILED' in line:
                log.warning('Authentication for Pilar failed.')
                self._logged_in = False

            else:
                # loop all commands and parse line
                commands_to_delete = []
                for cmd in self._commands:
                    # parse line
                    cmd.parse(line)

                    # is command finished?
                    if cmd.completed:
                        # remove it from list
                        commands_to_delete.append(cmd)

                # delete finished commands
                for cmd in commands_to_delete:
                    self._commands.remove(cmd)

    def connection_lost(self, exc):
        """ Called, when connection to server is lost. """
        self._loop.stop()

    def execute(self, command):
        # get next id
        self._id += 1

        # create command
        cmd = PilarCommand(command)
        cmd.id = self._id
        self._commands.append(cmd)

        # execute return
        cmd(self._transport)
        return cmd


class PilarDriver(object):
    """ Wrapper for easy communication with Pilar. """

    def __init__(self, host, port, username, password):
        """ Create new driver. """

        # init some stuff
        self._host = host
        self._port = port
        self._username = username
        self._password = password

        self._loop = None
        self._thread = None
        self._filters = []
        self.protocol = None
        self._closing_event = threading.Event()

        # errors
        self._has_error = False
        self._error_thread = None

    def open(self) -> bool:
        """ Open connection to SIImage. """

        # create and start thread
        self._thread = threading.Thread(target=self._thread_function, name='pilar')
        self._thread.start()

        # errors
        self._error_thread = threading.Thread(target=self._error_thread_func, name='pilarerr')
        self._error_thread.start()

        # wait for connection
        while self.protocol is None:
            time.sleep(0.1)

        # success
        return True

    def wait_for_login(self):
        while not self.protocol.logged_in:
            time.sleep(0.1)

    def _thread_function(self):
        """ Thread for asyncio event loop. """

        # create event loop
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        # create connection
        coro = self._loop.create_connection\
            (lambda: PilarClientProtocol(self, self._loop,
                                         self._username, self._password),
             self._host, self._port)

        # run event loop, until connection is made
        self._loop.run_until_complete(coro)

        # run loop forever
        self._loop.run_forever()

    def _error_thread_func(self):
        # run until closing
        while not self._closing_event.is_set():
            # not logged in?
            if self.protocol is None or not self.protocol.logged_in:
                self._closing_event.wait(5)

            # check for errors and clear them
            self._has_error = not self.clear_errors()

            # on error, wait and check
            if self._has_error:
                # wait a little
                self._closing_event.wait(5)

                # check again
                self._has_error = not self.check_errors()

            # wait five seconds
            self._closing_event.wait(5)

    @property
    def has_error(self):
        return self._has_error

    def close(self):
        """ Close connection to SIImage. """

        # safely close the connection
        self._closing_event.set()
        asyncio.run_coroutine_threadsafe(self.protocol.stop(), loop=self._loop)

    @property
    def is_open(self):
        """ Whether connection is open."""
        return self._protocol is not None

    def get(self, key):
        cmd = self.protocol.execute('GET ' + key)
        cmd.wait()
        return cmd.data

    def get_multi(self, keys):
        # join keys with ";" and execute
        cmd = self.protocol.execute('GET ' + ';'.join(keys))
        cmd.wait()
        return cmd.values

    def set(self, key, value, wait: bool = True, timeout: int = 5000, abort_event: threading.Event = None):
        """Set a variable with a given value.

        Args:
            key: Name of variable to set.
            value: New value.
            wait: Whether or not to wait for command.
            timeout: Timeout for waiting.
            abort_event: When set, wait is aborted.
        """

        # execute SET command
        cmd = self.protocol.execute('SET ' + key + '=' + str(value))

        # want to wait?
        if wait:
            cmd.wait(timeout=timeout, abort_event=abort_event)
            return cmd.error is None

        # return cmd
        return cmd

    def list_errors(self):
        """Fetch list of errors from telescope.

        From OpenTSI documentation about TELESCOPE.STATUS.LIST:
            A comma separated list of function groups that currently have
            problems in the following format:
            <group>|<level>[:<component>|<level>[;<component>...]]
            [:<error>|<detail>|<level>|<component>[;<error>...]]
            [,<group>...]
            <group> One of the above listed function groups
            <level> Bitwise “OR” of all errors in the group resp. compo-
            nent or for the individual error. The bits have the same
            meaning as for GLOBAL.
            <component> The OpenTCI module name (possibly includ-
            ing a index in []).
            <error> The hardware specific error code
            <detail> The hardware specific detail information for the
            error code.
            The information from <error>/<detail> should only be used
            for logging, as it is hardware specific and may change at any
            time.
            At most one entry per group is generated. If the delimiters
            should occur within the names or messages, they will be either
            escaped with a backslash or the entire entry is put in double
            quotes.
        """

        # init error list
        error_list = []

        # get list of errors
        errors = self.get('TELESCOPE.STATUS.LIST')

        # divide into groups and loop them
        for group in errors.split(','):
            # find last colon and split everything after that at semicolon
            errors = group[group.rfind(':')+1:].split(';')

            # loop errors
            for error in errors:
                # split by | to get name of error
                name = error.split('|')[0]
                if len(name) == 0:
                    continue

                # create error and add to list
                err = PilarError.create(name)
                error_list.append(err)

        # return list of errors
        return error_list

    def clear_errors(self):
        """Clears Pilar errors."""

        # get telescope status
        level = int(self.get('TELESCOPE.STATUS.GLOBAL'))

        # check level
        if level & (0x01 | 0x02):
            log.error('Found severe errors with level %d.', level)
        else:
            return True
            #log.info('Current error level is %d.', level)

        # check fatal errors
        self.list_errors()
        if PilarError.check_fatal():
            log.error('Cannot clear errors, fatal condition.')
            return False

        # do clearing
        log.info('Clearing telescope errors...')
        self.set('TELESCOPE.STATUS.CLEAR', level)

    def check_errors(self):
        """Check for errors after clearing."""

        # get telescope status
        level = int(self.get('TELESCOPE.STATUS.GLOBAL'))

        # check level
        if level & (0x01 | 0x02):
            log.error('Could not clear severe errors.')
            return False
        else:
            log.info('Remaining error level is %d.', level)
            return True

    def init(self, attempts: int = 3, wait: float = 10., attempt_timeout: float = 30.):
        """Initialize telescope

        Args:
            attempts (int): Number of attempts for initializing telescope.
            wait (float):   Wait time in seconds after sending command.
            attempt_timeout (float): Number of seconds to allow for each attempt.

        Returns:

        """

        # check, whether telescope is initialized already
        log.info("Querying ready status of telescope.")
        ready = self.get('TELESCOPE.READY_STATE')
        if float(ready) == 1.:
            log.info("Telescope already initialized.")
            return True

        # we give it a couple of attempts
        log.info("Initializing telescope...")
        for attempt in range(attempts):
            # init telescope
            self.set('TELESCOPE.READY', 1)

            # sleep a little
            time.sleep(wait)

            # wait for init
            waited = 0.
            while waited < attempt_timeout:
                # get status
                ready = self.get('TELESCOPE.READY_STATE')
                if float(ready) == 1.:
                    log.info("Telescope initialized.")
                    return True

                # sleep  a little
                waited += 0.5
                time.sleep(0.5)

        # we should never arrive here
        log.error('Could not initialize telescope.')
        return False

    def park(self, attempts: int = 3, wait: float = 10., attempt_timeout: float = 30.):
        # check, whether telescope is parked already
        log.info("Querying ready status of telescope.")
        ready = self.get('TELESCOPE.READY_STATE')
        if float(ready) == 0.:
            log.info("Telescope already parked.")
            return True

        # we give it a couple of attempts
        log.info("Parking telescope...")
        for attempt in range(attempts):
            # parking telescope
            self.set('TELESCOPE.READY', 0)

            # sleep a little
            time.sleep(wait)

            # wait for init
            waited = 0.
            while waited < attempt_timeout:
                # get status
                ready = self.get('TELESCOPE.READY_STATE')
                if float(ready) == 0.:
                    log.info("Telescope parked.")
                    return True

        # we should never arrive here
        log.error('Could not park telescope.')
        return False

    def reset_focus_offset(self):
        # get focus and offset
        focus = float(self.get('POSITION.INSTRUMENTAL.FOCUS.TARGETPOS'))
        offset = float(self.get('POSITION.INSTRUMENTAL.FOCUS.OFFSET'))

        # need to do something?
        if abs(offset) > 1e-5:
            # set new
            cmd1 = self.set('POSITION.INSTRUMENTAL.FOCUS.TARGETPOS', focus + offset, wait=False)
            cmd2 = self.set('POSITION.INSTRUMENTAL.FOCUS.OFFSET', 0, wait=False)

            # wait for both
            self.wait_for_all([cmd1, cmd2])

    @staticmethod
    def wait_for_all(commands):
        [cmd.wait() for cmd in commands]

    def focus(self, position, timeout=30000, accuracy=0.01, sleep=500, retry=3,
              sync_thermal=False, sync_port=False, sync_filter=False, disable_tracking=False,
              abort_event: threading.Event=None) -> bool:
        # reset any offset
        self.reset_focus_offset()

        # set sync_mode, first bit is always set
        sync_mode = 1
        if sync_thermal:
            log.info('Enabling synchronization with thermal model...')
            sync_mode |= 1 << 1
        if sync_port:
            log.info('Enabling synchronization with port specific offset...')
            sync_mode |= 1 << 2
        if sync_filter:
            log.info('Enabling synchronization with filter specific offset...')
            sync_mode |= 1 << 3
        if disable_tracking:
            log.info('Turning off focus motor during tracking...')
            sync_mode |= 1 << 4

        # setting new focus
        log.info('Setting new focus value to %.3fmm...', position)
        self.set('POINTING.SETUP.FOCUS.SYNCMODE', sync_mode)
        #self.set('POSITION.INSTRUMENTAL.FOCUS.OFFSET', 0)
        self.set('POINTING.SETUP.FOCUS.POSITION', position)
        self.set('POINTING.TRACK', 4)

        # loop until finished
        delta = 1e10
        waited = 0
        attempts = 0
        while delta >= accuracy:
            # abort?
            if abort_event is not None and abort_event.is_set():
                return False

            # sleep a little
            time.sleep(sleep / 1000.)
            waited += sleep

            # get focus distance
            delta = abs(float(self.get('POSITION.INSTRUMENTAL.FOCUS.TARGETDISTANCE')))
            log.info('Distance to new focus: %.3fmm.', delta)

            # waiting too long?
            if waited > timeout:
                # got more retries?
                if attempts < retry:
                    # yes, so try again
                    attempts += 1
                    waited = 0
                    log.warning('Focus timeout, starting attempt %d.', attempts+1)
                    self.set('POINTING.SETUP.FOCUS.POSITION', position)
                    self.set('POINTING.TRACK', 4)

                else:
                    # no, we're out of time
                    log.error('Focusing not possible.')
                    return False

        # get new focus
        foc = self.get('POSITION.INSTRUMENTAL.FOCUS.REALPOS')
        log.info('New focus position reached: %.3fmm.', float(foc))
        return True

    def goto(self, alt, az, abort_event: threading.Event) -> bool:
        log.info('Setting new position to Alt=%.4f, Az=%.4f...', alt, az)

        # stop telescope
        self.set('POINTING.TRACK', 0)

        # prepare derotator
        self.set('POINTING.SETUP.DEROTATOR.SYNCMODE', 3)
        self.set('POINTING.SETUP.DEROTATOR.OFFSET', 0.)

        # set new coordinates
        self.set('OBJECT.HORIZONTAL.AZ', az)
        self.set('OBJECT.HORIZONTAL.ALT', alt)

        # start moving
        self.set('POINTING.TRACK', 2)

        # wait for it
        for attempt in range(5):
            # abort?
            if abort_event.is_set():
                return False

            # wait
            success = self._wait_for_value('TELESCOPE.MOTION_STATE', '11', '0', abort_event=abort_event)
            if success:
                break

            # sleep a little and try again
            abort_event.wait(1)
            self.set('POINTING.TRACK', 2)
            log.warning('Attempt %d for moving to position failed.', attempt + 1)

        # success
        return success

    def track(self, ra, dec, abort_event: threading.Event) -> bool:
        log.info('Setting new position to RA=%.4f, Dec=%.4f...', ra, dec)

        # stop telescope
        self.set('POINTING.TRACK', 0)

        # prepare derotator
        self.set('POINTING.SETUP.DEROTATOR.SYNCMODE', 3)
        self.set('POINTING.SETUP.DEROTATOR.OFFSET', 0.)

        # set new coordinates
        self.set('OBJECT.EQUATORIAL.RA', ra/15.)
        self.set('OBJECT.EQUATORIAL.DEC', dec)

        # start tracking
        self.set('POINTING.TRACK', 1)

        # wait for it
        for attempt in range(5):
            # abort?
            if abort_event.is_set():
                return False

            # wait
            success = self._wait_for_value('TELESCOPE.MOTION_STATE', '11', '0', abort_event=abort_event)
            if success:
                break

            # got any errors?
            if len(self.list_errors()) > 0:
                return False

            # sleep a little and try again
            abort_event.wait(1)
            self.set('POINTING.TRACK', 2)
            log.warning('Attempt %d for moving to position failed.', attempt + 1)

        # success
        return success

    def _wait_for_value(self, var, value, not_value=None, abort_event: threading.Event = None):
        # sleep a little
        time.sleep(0.5)

        while True:
            # abort?
            if abort_event.is_set():
                return False

            # get variable
            val = float(self.get(var))

            # check
            if val == float(value):
                return True
            elif not_value is not None and val == float(not_value):
                return False

            # sleep a little
            time.sleep(1)

    def fits_data(self):
        return {
            'TEL-T1': float(self.get('AUXILIARY.SENSOR[3].VALUE')),
            'TEL-T2': float(self.get('AUXILIARY.SENSOR[1].VALUE')),
            'TEL-T3': float(self.get('AUXILIARY.SENSOR[2].VALUE')),
            'TEL-T4': float(self.get('AUXILIARY.SENSOR[4].VALUE')),
            'TEL-FOCU': float(self.get('POSITION.INSTRUMENTAL.FOCUS.REALPOS'))
        }

    def init_filters(self):
        # get number of filters
        num = int(self.get('TELESCOPE.CONFIG.PORT[2].FILTER'))
        log.info("Found %d filters.", num)

        # loop all filters
        self._filters = []
        for i in range(num):
            # set filter
            self.set('POINTING.SETUP.FILTER.INDEX', i)

            # get filter name
            name = self.get('POINTING.SETUP.FILTER.NAME')

            # strip quotes
            name = name.replace('"', '')

            # append to list
            log.info('Found filter %s.', name)
            self._filters.append(name)

    def filters(self):
        if not self._filters:
            self.init_filters()
        return self._filters

    def change_filter(self, filter_name, force_forward: bool = True, abort_event: threading.Event = None):
        # get current filter id
        cur_id = int(float(self.get('POSITION.INSTRUMENTAL.FILTER[2].CURRPOS')))

        # find ID of filter
        filter_id = self._filters.index(filter_name)
        if filter_id == cur_id:
            return True
        log.info('Changing to filter %s with ID %d.', filter_name, filter_id)

        # force only forward motion? if new ID is smaller than current one, first move to last filter
        if force_forward:
            # do until we're at the current filter
            while cur_id != filter_id:
                # how far can we go?
                for i in range(3):
                    # increase cur filter by one, wrap at end
                    cur_id += 1
                    if cur_id >= len(self._filters):
                        cur_id = 0

                    # got it?
                    if cur_id == filter_id:
                        break

                # move there
                if not self._change_filter_to_id(cur_id, abort_event):
                    log.info('Could not change to filter.')
                    return False

            # finished
            log.info('Successfully changed to filter %s.', filter_name)
            return True

        else:
            # simply go to requested filter
            if self._change_filter_to_id(filter_id, abort_event):
                log.info('Successfully changed to filter %s.', self._filters[filter_id])
                return True
            else:
                log.info('Could not change filter.')
                return False

    def _change_filter_to_id(self, filter_id: int, abort_event: threading.Event = None):
        # set it
        self.set('POINTING.SETUP.FILTER.INDEX', filter_id)
        self.set('POINTING.TRACK', 3)

        # wait for it
        return self._wait_for_value('POSITION.INSTRUMENTAL.FILTER[2].CURRPOS', filter_id, abort_event=abort_event)

    def filter_name(self, filter_id: int=None):
        if filter_id is None:
            filter_id = float(self.get('POSITION.INSTRUMENTAL.FILTER[2].CURRPOS'))
        return self._filters[int(filter_id)]



import datetime
import logging


log = logging.getLogger(__name__)


def create_error(ignore: bool = False, reset_max: int = 10, reset_timeout: int = 300,
                 accum_max: int = 5, accum_span: int = 86400) -> dict:
    """Creates an error

    Args:
        ignore:         If true, this error is ignored.
        reset_max:      Maximum number of resets.
        reset_timeout:  Minimum time between errors in seconds.
        accum_max:      Maximum number of errors to occur during accum_span.
        accum_span:     Accumulation time span in seconds.

    Returns:
        (dict) Dictionary containing values
    """
    return {
        'ignore': ignore,
        'reset_max': reset_max,
        'reset_timeout': reset_timeout,
        'accum_max': accum_max,
        'accum_span': accum_span
    }


# List of errors
ERRORS = {
    'ERR_GPS_PositionLost': create_error(ignore=True),
    'ERR_GPS_LeapSecond': create_error(ignore=True),
    'ERR_FilterWheel_RefMismatch': create_error(ignore=False, reset_timeout=0, accum_max=5, accum_span=3600),
    'ERR_Elevation_ETELExecError': create_error(ignore=False, reset_timeout=0, accum_max=5, accum_span=3600),
    'ERR_Oil_TemperatureLow': create_error(ignore=True)
}


class PilarError(object):
    _errors = {}

    def __init__(self, name):
        """Initializes an error

        Args:
            name:           Name of error.
        """
        self._name = name
        self._behaviour = ERRORS[name]
        self._dates = []

    @property
    def name(self):
        return self._name

    @property
    def ignore(self):
        return self._behaviour['ignore']

    @property
    def reset_max(self):
        return self._behaviour['reset_max']

    @property
    def reset_timeout(self):
        return self._behaviour['reset_timeout']

    @property
    def accum_max(self):
        return self._behaviour['accum_max']

    @property
    def accum_span(self):
        return self._behaviour['accum_span']

    def occur(self):
        """Should be called whenever error occurs.

        Returns:
            (bool) Whether or not it is safe to continue.
        """

        # ignore it?
        if self.ignore:
            return True

        # add now to list of dates
        self._dates.append(datetime.datetime.utcnow())

    def fatal(self) -> bool:
        # maximum number reached?
        if len(self._dates) > self.reset_max:
            log.warning('Error %s occurred more often than allowed (%d times).', self._name, len(self._dates))
            return True

        # time between last two errors?
        if len(self._dates) >= 2:
            # calculate difference between last two errors
            diff = (self._dates[-1] - self._dates[-2]).total_seconds()
            # check
            if diff < self.reset_timeout:
                log.warning('Duration (%.0fs) between last two errors of %s was shorter than allowed.',
                                diff, self._name)
                return True

        # collect errors within the last accum_span
        accum_errors = [d for d in self._dates if (self._dates[-1] - d).total_seconds() < self.accum_span]
        if len(accum_errors) > self.accum_max:
            log.warning('Too many (%d) errors of %s occurred during last %d seconds.',
                            len(accum_errors), self._name, self.accum_span)
            return True

        # everything okay
        return False

    @staticmethod
    def create(error_name: str) -> bool:
        # error not found?
        if error_name not in ERRORS:
            log.error('Unknown error "%s" occurred.', error_name)
            return

        # TODO: add more info
        log.error('Error "%s" occurred.', error_name)

        # did we encounter this error before?
        if error_name not in PilarError._errors:
            # create new one and add to list
            err = PilarError(error_name)
            PilarError._errors[error_name] = err

        # get error and let error occur
        err = PilarError._errors[error_name]
        err.occur()

        # return error
        return err

    @staticmethod
    def check_fatal() -> bool:
        # loop all errors
        for err in PilarError._errors.values():
            if err.fatal():
                return True

        # no fatal condition found
        return False



import logging
import threading
import time
from threading import Lock

from pyobs.events import FilterChangedEvent
from pyobs.interfaces import IFilters, IFitsHeaderProvider, IFocuser, ITemperatures, IAltAzMount
from pyobs.modules import timeout
from pyobs.modules.telescope.basetelescope import BaseTelescope
from pyobs.utils.threads import LockWithAbort
from .pilardriver import PilarDriver

log = logging.getLogger(__name__)


class PilarTelescope(BaseTelescope, IAltAzMount, IFilters, IFitsHeaderProvider, IFocuser, ITemperatures):
    def __init__(self, host: str, port: int, username: str, password: str, pilar_fits_headers: dict = None,
                 temperatures: dict = None, force_filter_forward: bool = True, *args, **kwargs):
        BaseTelescope.__init__(self, *args, **kwargs)

        # add thread func
        self._add_thread_func(self._pilar_update, True)

        # init pilar
        log.info('Connecting to Pilar at %s:%d...', host, port)
        self._pilar = PilarDriver(host, port, username, password)
        self._pilar.open()
        self._pilar.wait_for_login()

        # get list of filters
        self._filters = self._pilar.filters()
        self._force_filter_forward = force_filter_forward

        # get Pilar variables for status updates...
        self._pilar_variables = [
            'OBJECT.EQUATORIAL.RA', 'OBJECT.EQUATORIAL.DEC',
            'POSITION.EQUATORIAL.RA_J2000', 'POSITION.EQUATORIAL.DEC_J2000',
            'POSITION.HORIZONTAL.ZD', 'POSITION.HORIZONTAL.ALT', 'POSITION.HORIZONTAL.AZ',
            'POSITION.INSTRUMENTAL.FOCUS.REALPOS',
            'POSITION.INSTRUMENTAL.FILTER[2].CURRPOS',
            'POSITION.INSTRUMENTAL.DEROTATOR[2].REALPOS', 'POINTING.SETUP.DEROTATOR.OFFSET',
            'TELESCOPE.READY_STATE', 'TELESCOPE.MOTION_STATE',
            'POSITION.INSTRUMENTAL.AZ.OFFSET', 'POSITION.INSTRUMENTAL.ZD.OFFSET'
        ]

        # ... and add user defined ones
        self._pilar_fits_headers = pilar_fits_headers if pilar_fits_headers else {}
        for var in self._pilar_fits_headers.keys():
            if var not in self._pilar_variables:
                self._pilar_variables.append(var)

        # ... and temperatures
        self._temperatures = temperatures if temperatures else {}
        for var in self._temperatures.values():
            if var not in self._pilar_variables:
                self._pilar_variables.append(var)

        # create update thread
        self._status = {}
        self._lock = Lock()

        # optimal focus
        self._last_focus_time = None

        # some multi-threading stuff
        self._lock_focus = threading.Lock()
        self._abort_focus = threading.Event()
        self._lock_filter = threading.Lock()
        self._abort_filter = threading.Event()

    def open(self):
        """Open module."""
        BaseTelescope.open(self)

        # subscribe to events
        if self.comm:
            self.comm.register_event(FilterChangedEvent)

    def close(self):
        BaseTelescope.close(self)

        log.info('Closing connection to Pilar...')
        self._pilar.close()
        log.info('Shutting down...')

    def _pilar_update(self):
        # log
        log.info('Starting Pilar update thread...')

        while not self.closing.is_set():
            # do nothing on error
            if self._pilar.has_error:
                self.closing.wait(10)
                return

            # define values to request
            keys = self._pilar_variables

            # get data
            try:
                multi = self._pilar.get_multi(keys)
            except TimeoutError:
                # sleep a little and continue
                log.error('Request to Pilar timed out.')
                self.closing.wait(60)
                continue

            # set status
            with self._lock:
                self._status = {}
                for key in keys:
                    try:
                        self._status[key] = float(multi[key])
                    except ValueError:
                        # ignore it
                        pass

            # set motion status
            if float(self._status['TELESCOPE.READY_STATE']) == 0.:
                self._change_motion_status(BaseTelescope.Status.PARKED)
            elif 0. < float(self._status['TELESCOPE.READY_STATE']) < 1.:
                self._change_motion_status(BaseTelescope.Status.INITIALIZING)
            elif float(self._status['TELESCOPE.READY_STATE']) < 0.:
                self._change_motion_status(BaseTelescope.Status.ERROR)
            else:
                # telescope is initialized, check motion state
                ms = int(self._status['TELESCOPE.MOTION_STATE'])
                if ms & (1 << 0):
                    # first bit indicates moving
                    self._change_motion_status(BaseTelescope.Status.SLEWING)
                elif ms & (1 << 2):
                    # third bit indicates tracking
                    self._change_motion_status(BaseTelescope.Status.TRACKING)
                else:
                    # otherwise we're idle
                    self._change_motion_status(BaseTelescope.Status.IDLE)

            # sleep a second
            self.closing.wait(1)

        # log
        log.info('Shutting down Pilar update thread...')

    def get_fits_headers(self, *args, **kwargs) -> dict:
        # get headers from base
        hdr = BaseTelescope.get_fits_headers(self)

        # define values to request
        keys = {
            'TEL-FOCU': ('POSITION.INSTRUMENTAL.FOCUS.REALPOS', 'Focus position [mm]'),
            'TEL-ROT': ('POSITION.INSTRUMENTAL.DEROTATOR[2].REALPOS', 'Derotator instrumental position at end [deg]'),
            'DEROTOFF': ('POINTING.SETUP.DEROTATOR.OFFSET', 'Derotator offset [deg]'),
            'AZOFF': ('POSITION.INSTRUMENTAL.AZ.OFFSET', 'Azimuth offset'),
            'ALTOFF': ('POSITION.INSTRUMENTAL.ZD.OFFSET', 'Altitude offset')
        }

        # add ones from config
        for var, h in self._pilar_fits_headers.items():
            keys[h[0]] = (var, h[1])

        # Monet/S: 3=T1, 1=T2
        # Monet/N: 2=T1, 1=T2

        # create dict and add alt and filter
        with self._lock:
            status = self._status.copy()
        for key, var in keys.items():
            if var[0] in status:
                hdr[key] = (status[var[0]], var[1])

        # negative ALTOFF
        hdr['ALTOFF'] = (-hdr['ALTOFF'][0], hdr['ALTOFF'][1])

        # filter
        if 'POSITION.INSTRUMENTAL.FILTER[2].CURRPOS' in status:
            filter_id = status['POSITION.INSTRUMENTAL.FILTER[2].CURRPOS']
            hdr['FILTER'] = (self._pilar.filter_name(filter_id), 'Current filter')

        # return it
        return hdr

    def get_radec(self) -> (float, float):
        """Returns current RA and Dec.

        Returns:
            Tuple of current RA and Dec in degrees.
        """

        # check error
        if self._pilar.has_error:
            return None, None

        # get RA/Dec
        with self._lock:
            return self._status['POSITION.EQUATORIAL.RA_J2000'] * 15., self._status['POSITION.EQUATORIAL.DEC_J2000']

    def get_altaz(self) -> (float, float):
        """Returns current Alt and Az.

        Returns:
            Tuple of current Alt and Az in degrees.
        """

        # check error
        if self._pilar.has_error:
            return None, None

        # get Alt/Az
        with self._lock:
            return self._status['POSITION.HORIZONTAL.ALT'], self._status['POSITION.HORIZONTAL.AZ']

    def list_filters(self, *args, **kwargs) -> list:
        """List available filters.

        Returns:
            List of available filters.
        """
        return self._filters

    def get_filter(self, *args, **kwargs) -> str:
        """Get currently set filter.

        Returns:
            Name of currently set filter.
        """
        return self._pilar.filter_name()

    @timeout(60000)
    def set_filter(self, filter_name: str, *args, **kwargs):
        """Set the current filter.

        Args:
            filter_name: Name of filter to set.

        Raises:
            ValueError: If binning could not be set.
            AcquireLockFailed: If current motion could not be aborted.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # acquire lock
        with LockWithAbort(self._lock_filter, self._abort_filter):
            log.info('Changing filter to %s...', filter_name)
            self._pilar.change_filter(filter_name, force_forward=self._force_filter_forward,
                                      abort_event=self._abort_filter)
            log.info('Filter changed.')

            # send event
            self.comm.send_event(FilterChangedEvent(filter_name))

    def _move_altaz(self, alt: float, az: float, abort_event: threading.Event):
        """Actually moves to given coordinates. Must be implemented by derived classes.

        Args:
            alt: Alt in deg to move to.
            az: Az in deg to move to.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            Exception: On error.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # reset offsets
        self._reset_offsets()

        # start tracking
        log.info('Starting tracking at Alt=%.2f, Az=%.5f', alt, az)
        success = self._pilar.goto(alt, az, abort_event=abort_event)

        # finished
        if success:
            log.info('Reached destination.')
        else:
            raise ValueError('Could not reach destination.')

    def _track_radec(self, ra: float, dec: float, abort_event: threading.Event):
        """Actually starts tracking on given coordinates. Must be implemented by derived classes.

        Args:
            ra: RA in deg to track.
            dec: Dec in deg to track.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            Exception: On any error.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # reset offsets
        self._reset_offsets()

        # start tracking
        log.info('Starting tracking at RA=%.5f, Dec=%.5f', ra, dec)
        success = self._pilar.track(ra, dec, abort_event=abort_event)

        # finished
        if success:
            log.info('Reached destination.')
        else:
            raise ValueError('Could not reach destination.')

    def _reset_offsets(self):
        """Reset Alt/Az offsets."""
        self._pilar.set('POSITION.INSTRUMENTAL.ZD.OFFSET', 0)
        self._pilar.set('POSITION.INSTRUMENTAL.AZ.OFFSET', 0)

    def get_focus(self, *args, **kwargs) -> float:
        """Return current focus.

        Returns:
            Current focus.
        """
        return float(self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.CURRPOS'))

    def get_focus_offset(self, *args, **kwargs) -> float:
        """Return current focus offset.

        Returns:
            Current focus offset.
        """
        return float(self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.OFFSET'))

    @timeout(30000)
    def set_focus(self, focus: float, *args, **kwargs):
        """Sets new focus.

        Args:
            focus: New focus value.

        Raises:
            InterruptedError: If focus was interrupted.
            AcquireLockFailed: If current motion could not be aborted.
            TimeoutError: If focus could not be set in given time.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # acquire lock
        with LockWithAbort(self._lock_focus, self._abort_focus):
            # set focus
            log.info('Setting focus to %.4f...', focus)
            #self._pilar.set('POSITION.INSTRUMENTAL.FOCUS.TARGETPOS', focus,
            #                timeout=30000, abort_event=self._abort_focus)
            self._pilar.focus(focus)
            log.info('Reached new focus of %.4f.', float(self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.CURRPOS')))

    @timeout(30000)
    def set_focus_offset(self, offset: float, *args, **kwargs):
        """Sets focus offset.

        Args:
            offset: New focus offset.

        Raises:
            InterruptedError: If focus was interrupted.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # acquire lock
        with LockWithAbort(self._lock_focus, self._abort_focus):
            # set focus
            log.info('Setting focus offset to %.2f...', offset)
            self._pilar.set('POSITION.INSTRUMENTAL.FOCUS.OFFSET', offset,
                            timeout=10000, abort_event=self._abort_focus)
            log.info('Reached new focus offset of %.2f.', float(self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.OFFSET')))

    def set_altaz_offsets(self, dalt: float, daz: float, *args, **kwargs):
        """Move an Alt/Az offset, which will be reset on next call of track.

        Args:
            dalt: Altitude offset in degrees.
            daz: Azimuth offset in degrees.

        Raises:
            ValueError: If offset could not be set.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # set offsets
        log.info('Moving offset of dAlt=%.3f", dAz=%.3f".', dalt * 3600., daz * 3600.)
        self._pilar.set('POSITION.INSTRUMENTAL.ZD.OFFSET', -dalt)
        self._pilar.set('POSITION.INSTRUMENTAL.AZ.OFFSET', daz)

    def get_altaz_offsets(self, *args, **kwargs) -> (float, float):
        """Get Alt/Az offset.

        Returns:
            Tuple with alt and az offsets.
        """

        # get current offsets and return then
        dalt = -float(self._pilar.get('POSITION.INSTRUMENTAL.ZD.OFFSET'))
        daz = float(self._pilar.get('POSITION.INSTRUMENTAL.AZ.OFFSET'))
        return dalt, daz

    @timeout(300000)
    def init(self, *args, **kwargs):
        """Initialize telescope.

        Raises:
            ValueError: If telescope could not be initialized.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        log.info('Initializing telescope...')
        if not self._pilar.init():
            raise ValueError('Could not initialize telescope.')

        log.info('Initializing filter wheel...')
        self.set_filter(self._filters[-1])
        self.set_filter('clear')

    @timeout(300000)
    def park(self, *args, **kwargs):
        """Park telescope.

        Raises:
            ValueError: If telescope could not be parked.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # reset all offsets
        self._reset_offsets()

        # park telescope
        log.info('Parking telescope...')
        if not self._pilar.park():
            raise ValueError('Could not park telescope.')

    def get_temperatures(self, *args, **kwargs) -> dict:
        """Returns all temperatures measured by this module.

        Returns:
            Dict containing temperatures.
        """

        # lock status
        with self._lock:
            # get all temperatures
            temps = {}
            for name, var in self._temperatures.items():
                temps[name] = self._status[var]

            # return it
            return temps

    def stop_motion(self, device: str = None, *args, **kwargs):
        """Stop the motion.

        Args:
            device: Name of device to stop, or None for all.
        """
        pass


__all__ = ['PilarTelescope']



from setuptools import setup, find_packages

# setup
setup(
    name='pyobs-pilar',
    version='0.8',
    description='pyobs component for the Pilar TCS',
    packages=['pyobs_pilar'],
    install_requires=[
        'astropy',
        'pyobs-core'
    ]
)
